from .common import CreateResource, ListResources, GetResource, StartResource
from .. import serializers
from ..repositories.experiments import ParseExperimentDictMixin, GetBaseExperimentApiUrlMixin


class CreateHyperparameterJob(GetBaseExperimentApiUrlMixin, CreateResource):
    SERIALIZER_CLS = serializers.HyperparameterSchema

    def get_request_url(self, **_):
        return "/hyperopt/"


class CreateAndStartHyperparameterJob(CreateHyperparameterJob):
    def get_request_url(self, **_):
        return "/hyperopt/create_and_start/"


class ListHyperparameterJobs(GetBaseExperimentApiUrlMixin, ParseExperimentDictMixin, ListResources):
    def get_request_url(self, **kwargs):
        return "/hyperopt/"

    def _parse_objects(self, data, **kwargs):
        experiments = []
        for experiment_dict in data["data"]:
            experiment_dict.update(experiment_dict["templateHistory"].get("params", {}))
            experiment = self._parse_object(experiment_dict)
            experiments.append(experiment)

        return experiments

    def _get_request_params(self, kwargs):
        return {"limit": -1}


class GetHyperparameterTuningJob(GetBaseExperimentApiUrlMixin, ParseExperimentDictMixin, GetResource):
    def get_request_url(self, **kwargs):
        id_ = kwargs["id"]
        url = "/hyperopt/{}/".format(id_)
        return url

    def _parse_object(self, job_dict, **kwargs):
        data = job_dict["data"]
        data.update(data["templateHistory"].get("params", {}))
        instance = super(GetHyperparameterTuningJob, self)._parse_object(data)
        return instance


class StartHyperparameterTuningJob(GetBaseExperimentApiUrlMixin, StartResource):
    VALIDATION_ERROR_MESSAGE = "Failed to start hyperparameter tuning job"

    def get_request_url(self, **kwargs):
        id_ = kwargs["id"]
        url = "/hyperopt/{}/start/".format(id_)
        return url



import marshmallow as ma

from .base import BaseSchema
from .. import models


class AutoscalingMetricSchema(BaseSchema):
    MODEL = models.AutoscalingMetric

    type = ma.fields.Str()
    name = ma.fields.Str()
    value_type = ma.fields.Str(dump_to="valueType", load_from="valueType")
    value = ma.fields.Float()


class AutoscalingDefinitionSchema(BaseSchema):
    MODEL = models.AutoscalingDefinition

    min_instance_count = ma.fields.Int(dump_to="minInstanceCount", load_from="minInstanceCount")
    max_instance_count = ma.fields.Int(dump_to="maxInstanceCount", load_from="maxInstanceCount")
    scale_cooldown_period = ma.fields.Int(dump_to="scaleCooldownPeriod", load_from="scaleCooldownPeriod")
    metrics = ma.fields.Nested(AutoscalingMetricSchema, many=True, default=None)


class DeploymentSchema(BaseSchema):
    MODEL = models.Deployment

    id = ma.fields.Str(dump_to="id", load_from="id")
    name = ma.fields.Str(required=True)
    endpoint = ma.fields.Str()
    api_type = ma.fields.Str(dump_to="apiType", load_from="apiType")

    state = ma.fields.Str()

    model_id = ma.fields.Str(rquired=True, dump_to="modelId", load_from="modelId")
    project_id = ma.fields.Str(dump_to="projectId", load_from="projectId")

    image_url = ma.fields.Str(required=True, dump_to="imageUrl", load_from="imageUrl")
    deployment_type = ma.fields.Str(required=True, dump_to="deploymentType", load_from="deploymentType")
    machine_type = ma.fields.Str(required=True, dump_to="machineType", load_from="machineType")
    instance_count = ma.fields.Int(required=True, dump_to="instanceCount", load_from="instanceCount")
    container_model_path = ma.fields.Str(dump_to="containerModelPath", load_from="containerModelPath")
    image_username = ma.fields.Str(dump_to="imageUsername", load_from="imageUsername")
    image_password = ma.fields.Str(dump_to="imagePassword", load_from="imagePassword")
    image_server = ma.fields.Str(dump_to="imageServer", load_from="imageServer")
    container_url_path = ma.fields.Str(dump_to="containerUrlPath", load_from="containerUrlPath")
    endpoint_url_path = ma.fields.Str(dump_to="endpointUrlPath", load_from="endpointUrlPath")
    method = ma.fields.Str(dump_to="method", load_from="method")
    docker_args = ma.fields.List(ma.fields.Str(), dump_to="dockerArgs", load_from="dockerArgs")
    env = ma.fields.Dict(dump_to="env", load_from="env")
    ports = ma.fields.Str(dump_to="ports", load_from="ports")
    auth_username = ma.fields.Str(dump_to="oauthKey", load_from="oauthKey")
    auth_password = ma.fields.Str(dump_to="oauthSecret", load_from="oauthSecret")
    cluster_id = ma.fields.Str(dump_to="clusterId", load_from="clusterId")
    tags = ma.fields.List(ma.fields.Str(), load_only=True)
    command = ma.fields.Str()
    metrics_url = ma.fields.Str(dump_to="metricsURL", load_from="metricsURL")
    workspace_url = ma.fields.Str(dump_to="workspaceUrl", load_from="workspaceUrl")
    workspace_ref = ma.fields.Str(dump_to="workspaceRef", load_from="workspaceRef")
    workspace_username = ma.fields.Str(dump_to="workspaceUsername", load_from="workspaceUsername")
    workspace_password = ma.fields.Str(dump_to="workspacePassword", load_from="workspacePassword")
    dt_created = ma.fields.DateTime(dump_to="dtCreated", load_from="dtCreated")
    dt_modified = ma.fields.DateTime(dump_to="dtModified", load_from="dtModified")
    dt_started = ma.fields.DateTime(dump_to="dtStarted", load_from="dtStarted")
    dt_stopped = ma.fields.DateTime(dump_to="dtStopped", load_from="dtStopped")
    dt_deleted = ma.fields.DateTime(dump_to="dtDeleted", load_from="dtDeleted")
    autoscaling = ma.fields.Nested(AutoscalingDefinitionSchema)


class DeploymentCreateSchema(DeploymentSchema):
    cluster_id = ma.fields.Str(dump_to="cluster", load_from="clusterId")



import os
import sys
import tempfile

import mock
from pytest import fixture

from gradient.api_sdk.s3_uploader import S3FileUploader, ExperimentFileUploader, S3ModelFileUploader, \
    ExperimentWorkspaceDirectoryUploader, S3ModelUploader

open_path = "builtins.open"
if sys.version_info[0] < 3:
    open_path = "__builtin__.open"


@fixture
def client():
    client = mock.MagicMock()
    client.get = mock.MagicMock()
    client.put = mock.MagicMock()
    client.post = mock.MagicMock()
    return client


@fixture
def file_uploader(client):
    mock_get_client = mock.MagicMock()
    mock_get_client.return_value = client

    mock_encoder_class = mock.Mock(spec=S3FileUploader.DEFAULT_MULTIPART_ENCODER_CLS)
    uploader = S3FileUploader(multipart_encoder_cls=mock_encoder_class)
    uploader._get_client = mock_get_client
    return uploader


def get_ps_exp_valid_response(mock_url):
    mock_response = mock.MagicMock()
    mock_response.status_code = 200
    mock_response.ok = True
    mock_response.json.return_value = {
        "message": "success",
        "data": {
            "url": mock_url,
            "bucket_name": "bucketname",
            "fields": {
                "key": "foo_key"
            }
        }
    }
    return mock_response


class TestS3FileUploader(object):
    def test_should_post_file_data_and_return_valid_url(self, client, file_uploader):
        upload_url = "s3://url"
        _mock_open = mock.mock_open(read_data="data")
        with mock.patch(open_path, _mock_open) as mock_file:
            file_uploader.upload("filename", upload_url, {"key": "foo"})

        file_uploader._get_client.assert_called_with(upload_url)
        client.post.assert_called()


class TestExperimentFileUploader(object):
    def test_should_post_file_data_and_return_valid_url(self, client, file_uploader):
        upload_url = "s3://url"
        mock_response = get_ps_exp_valid_response(upload_url)

        mock_api_client = mock.MagicMock()
        mock_api_client.get.return_value = mock_response

        uploader = ExperimentFileUploader("api_key", uploader=file_uploader)
        uploader.experiments_api = mock_api_client

        _mock_open = mock.mock_open(read_data="data")
        with mock.patch(open_path, _mock_open) as mock_file:
            uploader.upload("foo", "pjHandle", "clHandle")

        file_uploader._get_client.assert_called_with(upload_url)
        client.post.assert_called()


class TestS3ModelFileUploader(object):
    def test_should_post_file_data_and_return_valid_url(self, client, file_uploader):
        upload_url = "s3://url"

        mock_response = mock.MagicMock()
        mock_response.status_code = 200
        mock_response.ok = True
        mock_response.json.return_value = upload_url

        mock_api_client = mock.MagicMock()
        mock_api_client.get.return_value = mock_response

        uploader = S3ModelFileUploader("api_key", s3uploader=file_uploader)
        uploader.ps_api_client = mock_api_client
        _mock_open = mock.mock_open(read_data="data")

        with mock.patch(open_path, _mock_open) as mock_file:
            uploader.upload("foo", "mdHandle")

        file_uploader._get_client.assert_called_with(upload_url)
        client.post.assert_called()


class TestS3ModelUploader(object):

    @mock.patch("os.path.isdir")
    @mock.patch("gradient.api_sdk.archivers.ZipArchiver.archive")
    def test_should_post_file_data_and_return_valid_url(self, archive_patched, is_dir_patched, client, file_uploader):
        upload_url = "s3://url"

        mock_response = mock.MagicMock()
        mock_response.status_code = 200
        mock_response.ok = True
        mock_response.json.return_value = upload_url

        mock_api_client = mock.MagicMock()
        mock_api_client.get.return_value = mock_response

        uploader = S3ModelUploader("api_key", s3uploader=file_uploader)
        uploader.ps_api_client = mock_api_client
        _mock_open = mock.mock_open(read_data="data")

        is_dir_patched.return_value = True

        with mock.patch(open_path, _mock_open) as mock_file:
            uploader.upload("foo", "mdHandle")

        archive_patched.assert_called_once_with("foo", os.path.join(tempfile.gettempdir(), "model.zip"))
        file_uploader._get_client.assert_called_with(upload_url)
        client.post.assert_called()


class TestExperimentWorkspaceDirectoryUploader(object):

    def test_should_post_file_data_and_return_valid_url(self, client, file_uploader):
        upload_url = "s3://url"

        mock_response = get_ps_exp_valid_response(upload_url)

        mock_api_client = mock.MagicMock()
        mock_api_client.get.return_value = mock_response
        project_uploader = ExperimentFileUploader("api_key", uploader=file_uploader)
        project_uploader.experiments_api = mock_api_client

        uploader = ExperimentWorkspaceDirectoryUploader(api_key="api_key", archiver=mock.MagicMock(),
                                                        project_uploader=project_uploader)

        _mock_open = mock.mock_open(read_data="data")
        with mock.patch(open_path, _mock_open) as mock_file:
            uploader.upload("foo", "mdHandle")

        file_uploader._get_client.assert_called_with(upload_url)
        client.post.assert_called()



import base64
import os
import random
import string
from collections import OrderedDict

import progressbar
import six
from requests_toolbelt.multipart import encoder

from . import constants, sdk_exceptions


class MessageExtractor(object):
    def get_message_from_response_data(self, response_data, sep="\n"):
        """
        :type response_data: None|dict|list|str
        :type sep: str
        :rtype: str
        """
        messages = list(self.get_error_messages(response_data))
        msg = sep.join(messages)

        return msg

    def get_error_messages(self, data, add_prefix=False):
        if isinstance(data, dict):
            for key, value in sorted(data.items()):
                if key in ("error", "errors", "message", "messages", "title", "description"):
                    for message in self.get_error_messages(value):
                        yield message

                # when key == "details" and value is a dict then values should be prefixed with key
                if add_prefix:
                    for message in self.get_error_messages(value):
                        # there is some useless message in data["context"]
                        if key != "context":
                            yield key + ": " + message
                if key == "details":
                    for message in self.get_error_messages(value, add_prefix=True):
                        yield message

        if isinstance(data, list):
            for element in data:
                for message in self.get_error_messages(element):
                    yield message

        if isinstance(data, six.string_types):
            yield data

        if six.PY3 and isinstance(data, bytes):
            yield data.decode('utf-8')


def print_dict_recursive(input_dict, logger, indent=0, tabulator="  "):
    for key, val in input_dict.items():
        logger.log("%s%s:" % (tabulator * indent, key))
        if type(val) is dict:
            print_dict_recursive(OrderedDict(val), logger, indent + 1)
        else:
            logger.log("%s%s" % (tabulator * (indent + 1), val))


class ExperimentsClientHelpersMixin(object):
    def _get_experiment_type_id(self, value):
        if isinstance(value, int):
            return value

        try:
            experiment_type_id = constants.MULTI_NODE_EXPERIMENT_TYPES_MAP[value]
        except KeyError as e:
            raise sdk_exceptions.GradientSdkError("Invalid experiment type: {}".format(e))

        return experiment_type_id


def validate_auth_options(auth_username, auth_password, generate_auth):
    if generate_auth and any((auth_username, auth_password)):
        raise sdk_exceptions.InvalidParametersError(
            "Use either auth_username and auth_password or generate_auth",
        )

    # checking if both or none auth parameters were used
    if len([val for val in (auth_username, auth_password) if not bool(val)]) == 1:
        raise sdk_exceptions.InvalidParametersError("auth_username and auth_password have to be used together")


def generate_credential(n):
    cred = ''.join(random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits) for _ in range(n))
    return cred


def generate_credentials_pair(n):
    username = generate_credential(n)
    password = generate_credential(n)
    return username, password


def base64_encode(s):
    if six.PY3:
        s = bytes(s, encoding="utf8")

    encoded_str = base64.b64encode(s)

    if six.PY3:  # Python3's base64.b64encode returns a bytes instance so it should be converted back to unicode
        encoded_str = encoded_str.decode("utf-8")

    return encoded_str


def base64_encode_attribute(data, name):
    raw = getattr(data, name)
    if raw is None:
        return
    encoded_value = base64_encode(raw)
    setattr(data, name, encoded_value)


def concatenate_urls(fst_part, snd_part):
    fst_part = fst_part if not fst_part.endswith("/") else fst_part[:-1]
    template = "{}{}" if snd_part.startswith("/") else "{}/{}"
    concatenated = template.format(fst_part, snd_part)
    return concatenated


class MultipartEncoder(object):
    def __init__(self, fields):
        mp_encoder = encoder.MultipartEncoder(fields=fields)
        self.monitor = encoder.MultipartEncoderMonitor(mp_encoder, callback=self._create_callback(mp_encoder))

    def get_monitor(self):
        return self.monitor

    @staticmethod
    def _create_callback(encoder_obj):
        pass


class MultipartEncoderWithProgressbar(MultipartEncoder):
    @staticmethod
    def _create_callback(encoder_obj):
        bar = progressbar.ProgressBar(max_value=encoder_obj.len)

        def callback(monitor):
            if monitor.bytes_read == bar.max_value:
                bar.finish()
            else:
                bar.update(monitor.bytes_read)

        return callback


class PathParser(object):
    LOCAL_DIR = 0
    LOCAL_FILE = 1
    GIT_URL = 2
    S3_URL = 3
    HTTP_URL = 4

    @classmethod
    def parse_path(cls, path):
        if path is None:
            return None

        if cls.is_local_dir(path):
            return cls.LOCAL_DIR

        if cls.is_local_zip_file(path):
            return cls.LOCAL_FILE

        if cls.is_git_url(path):
            return cls.GIT_URL

        if cls.is_s3_url(path):
            return cls.S3_URL

        if cls.is_http_url(path):
            return cls.HTTP_URL

        raise sdk_exceptions.WrongPathError("Given path is neither local path, nor valid URL")

    @staticmethod
    def is_local_dir(path):
        return os.path.exists(path) and os.path.isdir(path)

    @staticmethod
    def is_local_zip_file(path):
        return os.path.exists(path) and os.path.isfile(path) and path.endswith(".zip")

    @staticmethod
    def is_git_url(path):
        return not os.path.exists(path) and path.endswith(".git") or path.lower().startswith("git:")

    @staticmethod
    def is_s3_url(path):
        return not os.path.exists(path) and path.lower().startswith("s3:")

    @staticmethod
    def is_http_url(path):
        normalized_path = path.lower()
        return not os.path.exists(path) and (normalized_path.startswith("http:") or normalized_path.startswith("https:"))

    @classmethod
    def is_remote_path(cls, path):
        return cls.is_s3_url(path) or cls.is_git_url(path) or cls.is_http_url(path)

    @classmethod
    def is_local_path(cls, path):
        return cls.is_local_dir(path) or cls.is_local_zip_file(path)



import datetime
import functools
import json
import re

import click
import six
import termcolor
import yaml
from click.exceptions import Exit
from click_didyoumean import DYMMixin
from click_help_colors import HelpColorsGroup

from gradient.api_sdk.config import config
from gradient.cli import cli_types

OPTIONS_FILE_OPTION_NAME = "optionsFile"
OPTIONS_FILE_PARAMETER_NAME = "options_file"
OPTIONS_DUMP_FILE_OPTION_NAME = "createOptionsFile"


def del_if_value_is_none(dict_, del_all_falsy=False):
    """Remove all elements with value == None"""
    for key, val in list(dict_.items()):
        if val is None or (del_all_falsy and not val):
            del dict_[key]


def jsonify_dicts(dict_):
    json_fields = [
        "envVars",
        "nodeAttrs"
    ]
    for field in json_fields:
        if field in dict_:
            dict_[field] = json.dumps(dict_[field])


class ClickGroup(DYMMixin, HelpColorsGroup):
    pass


def deprecated(msg):
    deprecated_invoke_notice = msg + """\nFor more information, please see:

https://docs.paperspace.com
If you depend on functionality not listed there, please file an issue."""

    def new_invoke(self, ctx):
        click.echo(click.style(deprecated_invoke_notice, fg='red'), err=True)
        super(type(self), self).invoke(ctx)

    def decorator(f):
        f.invoke = functools.partial(new_invoke, f)

    return decorator


def get_option_name(options_strings):
    for opt in options_strings:
        if not opt.startswith("-"):
            return opt

        if opt.startswith("--"):
            return opt[2:]


class ReadValueFromConfigFile(click.Parameter):
    def handle_parse_result(self, ctx, opts, args):
        config_file = ctx.params.get(OPTIONS_FILE_PARAMETER_NAME)
        if self.should_read_value_from_file(opts, args, config_file):
            with open(config_file) as f:
                config_data = yaml.load(f, Loader=yaml.FullLoader)
                option_name = get_option_name(self.opts)
                if (isinstance(self, GradientObjectListOption) and
                        self.object_list_name in config_data):
                    value_list = []
                    for object_list_item in config_data[self.object_list_name]:
                        value_list.append(object_list_item.get(self.object_key))

                    if value_list:
                        opts[self.name] = value_list
                else:
                    value = config_data.get(option_name)
                    if value is not None:
                        if isinstance(value, dict):
                            value = json.dumps(value)
                        elif self.multiple and isinstance(value, six.string_types):
                            value = (value,)
                        # yaml.load turns datetime strings into datetime instances so we turn it back to a string
                        elif isinstance(value, datetime.datetime):
                            value = value.strftime("%Y-%m-%dT%H:%M:%S")

                        opts[self.name] = value

        rv = super(ReadValueFromConfigFile, self).handle_parse_result(
            ctx, opts, args)
        return rv

    def should_read_value_from_file(self, opts, args, config_file):
        """
        :rtype: bool
        """
        raise NotImplementedError


class ColorExtrasInCommandHelpMixin(object):
    def get_help_record(self, *args, **kwargs):
        rv = super(ColorExtrasInCommandHelpMixin,
                   self).get_help_record(*args, **kwargs)

        if not config.USE_CONSOLE_COLORS:
            return rv

        try:
            help_str = rv[1]
        except (IndexError, TypeError):
            return rv

        if help_str:
            help_str = self._color_extras(help_str)
            rv = rv[0], help_str
        return rv

    def _color_extras(self, s):
        pattern = re.compile(r"^.*(\[.*\])$")
        found = re.findall(pattern, s)
        if found:
            extras_str = found[-1]
            coloured_extras_str = self._color_str(extras_str)
            s = s.replace(extras_str, coloured_extras_str)

        return s

    def _color_str(self, s):
        s = termcolor.colored(s, config.HELP_HEADERS_COLOR)
        return s


class GradientArgument(ColorExtrasInCommandHelpMixin, ReadValueFromConfigFile, click.Argument):
    def should_read_value_from_file(self, opts, args, config_file):
        return opts.get(self.name) in (None, ()) and config_file


class GradientOption(ColorExtrasInCommandHelpMixin, ReadValueFromConfigFile, click.Option):
    def should_read_value_from_file(self, opts, args, config_file):
        return self.name not in opts and config_file


api_key_option = click.option(
    "--apiKey",
    "api_key",
    help="API key to use this time only",
    cls=GradientOption,
)


def generate_options_template(ctx, param, value):
    if not value:
        return value

    params = {}
    objects_value_lists = {}  # Object list name -> object key -> list of values
    for param in ctx.command.params:
        option_name = get_option_name(param.opts)
        if option_name in (OPTIONS_FILE_OPTION_NAME, OPTIONS_DUMP_FILE_OPTION_NAME):
            continue

        option_value = ctx.params.get(param.name) or param.default

        # If this is an object list type option, add its value list to
        # the specific object list set of value lists.
        if isinstance(param, GradientObjectListOption):
            new_value_list = option_value if option_value is not None else []
            object_list_name = param.object_list_name
            value_lists = objects_value_lists.setdefault(object_list_name, {})
            value_lists[param.object_key] = new_value_list
            continue

        if isinstance(param.type, cli_types.ChoiceType):
            for key, val in param.type.type_map.items():
                if val == option_value:
                    option_value = key

        params[option_name] = option_value

    # Transform value lists into objects
    object_lists = {}
    for object_list_name, value_lists in objects_value_lists.items():
        # Find maximum length value list and assume it lines up
        # with other object list options
        object_list = object_lists.setdefault(object_list_name, [])

        num_items = max(len(value_list) for value_list in value_lists.values())

        for i in range(num_items):
            new_object = {}
            for object_key, value_list in value_lists.items():
                if i < len(value_list):
                    new_object[object_key] = value_list[i]
                else:
                    new_object[object_key] = None

            object_list.append(new_object)

    # Add all object lists to overall params
    # for all non-empty object lists
    for object_list_name, object_list in object_lists.items():
        if object_list:
            params[object_list_name] = object_list

    with open(value, "w") as f:
        yaml.safe_dump(params, f, default_flow_style=False)

    raise Exit  # to stop execution without executing the command


def options_file(f):
    options = [
        click.option(
            "--" + OPTIONS_FILE_OPTION_NAME,
            OPTIONS_FILE_PARAMETER_NAME,
            is_eager=True,
            help="Path to YAML file with predefined options",
            type=click.Path(exists=True, resolve_path=True)
        ),
        click.option(
            "--" + OPTIONS_DUMP_FILE_OPTION_NAME,
            callback=generate_options_template,
            expose_value=False,
            help="Generate template options file",
            type=click.Path(writable=True, resolve_path=True)
        )
    ]
    return functools.reduce(lambda x, opt: opt(x), reversed(options), f)


class GradientObjectListOption(GradientOption):
    def __init__(self, object_name, param_decls, **kwargs):
        super(GradientObjectListOption, self).__init__(param_decls, **kwargs)
        self.object_name = object_name
        self.object_list_name = self.object_name + "s"
        self.object_key = self._compose_object_key()

    # Get this option's object key for the objects in the object list
    def _compose_object_key(self):
        option_name = get_option_name(self.opts)
        if not option_name.startswith(self.object_name):
            return option_name

        object_key = option_name[len(self.object_name):]
        object_key = object_key[0].lower() + object_key[1:]
        return object_key


class GradientDatasetOption(GradientObjectListOption):
    def __init__(self, param_decls, **kwargs):
        super(GradientDatasetOption, self).__init__("dataset", param_decls, **kwargs)


def validate_comma_split_option(comma_option_value, option_value, raise_if_no_values=False):
    if raise_if_no_values and not any((comma_option_value, option_value)):
        raise click.UsageError("No tags provided")

    if comma_option_value or option_value:
        if option_value:
            option_value = list(option_value)
        else:
            option_value = list()
        if comma_option_value:
            comma_option_value = [s.strip() for s in comma_option_value.split(",")]
            option_value.extend(comma_option_value)

        tags = sorted(set(option_value))
        return tags



import datetime

from .base_client import BaseClient, TagsSupportMixin
from .. import repositories, models, constants, utils
from ..sdk_exceptions import ResourceCreatingDataError, InvalidParametersError
from ..validation_messages import EXPERIMENT_MODEL_PATH_VALIDATION_ERROR


class ExperimentsClient(TagsSupportMixin, utils.ExperimentsClientHelpersMixin, BaseClient):
    entity = "experiment"

    def create_single_node(
            self,
            project_id,
            machine_type,
            command,
            ports=None,
            name=None,
            workspace_url=None,
            workspace_ref=None,
            workspace_username=None,
            workspace_password=None,
            datasets=None,
            working_directory=None,
            artifact_directory=None,
            cluster_id=None,
            experiment_env=None,
            model_type=None,
            model_path=None,
            is_preemptible=False,
            container=None,
            container_user=None,
            registry_username=None,
            registry_password=None,
            registry_url=None,
            tags=None,
    ):
        """
        Create single node experiment

        :param str name: Name of new experiment  [required]
        :param str project_id: Project ID  [required]
        :param str machine_type: Machine type [required]
        :param str command: Container entrypoint command  [required]
        :param str ports: Port to use in new experiment
        :param str workspace_url: Project git repository url
        :param str workspace_ref: Git commit hash, branch name or tag
        :param str workspace_username: Project git repository username
        :param str workspace_password: Project git repository password
        :param dict|list[dict]|tuple[dict] datasets: Dict or list of dicts describing dataset(s) used in experiment.
                                                     Required keys: "id" or "url"
                                                     Optional keys: "tag" for S3 tag and "auth" for S3 token
        :param str working_directory: Working directory for the experiment
        :param str artifact_directory: Artifacts directory
        :param str cluster_id: Cluster ID
        :param dict experiment_env: Environment variables in a JSON
        :param str model_type: defines the type of model that is being generated by the experiment. Model type must be one of Tensorflow, ONNX, or Custom
        :param str model_path: Model path
        :param bool is_preemptible: Is preemptible
        :param str container: Container (dockerfile) [required]
        :param str container_user: Container user for running the specified command in the container. If no containerUser is specified, the user will default to 'root' in the container.
        :param str registry_username: Registry username for accessing private docker registry container if necessary
        :param str registry_password: Registry password for accessing private docker registry container if necessary
        :param str registry_url: Registry server URL for accessing private docker registry container if necessary
        :param list[str] tags: List of tags

        :returns: experiment handle
        :rtype: str
        """
        self._validate_arguments(model_type=model_type, model_path=model_path)

        if not is_preemptible:
            is_preemptible = None

        datasets = self._dataset_dicts_to_instances(datasets)

        experiment = models.SingleNodeExperiment(
            experiment_type_id=constants.ExperimentType.SINGLE_NODE,
            name=name,
            project_id=project_id,
            machine_type=machine_type,
            ports=ports,
            workspace_url=workspace_url,
            workspace_ref=workspace_ref,
            workspace_username=workspace_username,
            workspace_password=workspace_password,
            datasets=datasets,
            working_directory=working_directory,
            artifact_directory=artifact_directory,
            cluster_id=cluster_id,
            experiment_env=experiment_env,
            model_type=model_type,
            model_path=model_path,
            is_preemptible=is_preemptible,
            container=container,
            command=command,
            container_user=container_user,
            registry_username=registry_username,
            registry_password=registry_password,
            registry_url=registry_url,
        )

        repository = self.build_repository(repositories.CreateSingleNodeExperiment)
        handle = repository.create(experiment)

        if tags:
            self.add_tags(entity_id=handle, tags=tags)

        return handle

    def create_multi_node(
            self,
            project_id,
            worker_container,
            worker_machine_type,
            worker_command,
            worker_count,
            parameter_server_container,
            parameter_server_machine_type,
            parameter_server_command,
            parameter_server_count,
            name=None,
            experiment_type_id=constants.ExperimentType.GRPC_MULTI_NODE,
            ports=None,
            workspace_url=None,
            workspace_ref=None,
            workspace_username=None,
            workspace_password=None,
            datasets=None,
            working_directory=None,
            artifact_directory=None,
            cluster_id=None,
            experiment_env=None,
            model_type=None,
            model_path=None,
            is_preemptible=False,
            worker_container_user=None,
            worker_registry_username=None,
            worker_registry_password=None,
            worker_registry_url=None,
            parameter_server_container_user=None,
            parameter_server_registry_username=None,
            parameter_server_registry_password=None,
            parameter_server_registry_url=None,
            tags=None,
    ):
        """
        Create multinode experiment

        :param str name: Name of new experiment  [required]
        :param str project_id: Project ID  [required]
        :param str worker_container: Worker container (dockerfile) [required]
        :param str worker_machine_type: Worker machine type  [required]
        :param str worker_command: Worker command  [required]
        :param int worker_count: Worker count  [required]
        :param str parameter_server_container: Parameter server container  [required]
        :param str parameter_server_machine_type: Parameter server machine type  [required]
        :param str parameter_server_command: Parameter server command  [required]
        :param int parameter_server_count: Parameter server count  [required]
        :param int|str experiment_type_id: Experiment Type ID
        :param str ports: Port to use in new experiment
        :param str workspace_url: Project git repository url
        :param str workspace_ref: Git commit hash, branch name or tag
        :param str workspace_username: Project git repository username
        :param str workspace_password: Project git repository password
        :param dict|list[dict]|tuple[dict] datasets: Dict or list of dicts describing dataset(s) used in experiment.
                                                     Required keys: "url"
                                                     Optional keys: "tag" for S3 tag and "auth" for S3 token
        :param str working_directory: Working directory for the experiment
        :param str artifact_directory: Artifacts directory
        :param str cluster_id: Cluster ID
        :param dict experiment_env: Environment variables in a JSON
        :param str model_type: defines the type of model that is being generated by the experiment. Model type must be one of Tensorflow, ONNX, or Custom
        :param str model_path: Model path
        :param bool is_preemptible: Is preemptible
        :param str worker_container_user: Worker container user
        :param str worker_registry_username: Registry username for accessing private docker registry container if necessary
        :param str worker_registry_password: Registry password for accessing private docker registry container if necessary
        :param str worker_registry_url: Registry server URL for accessing private docker registry container if necessary
        :param str parameter_server_container_user: Parameter server container user
        :param str parameter_server_registry_username: Registry username for accessing private docker registry container if necessary
        :param str parameter_server_registry_password: Registry password for accessing private docker registry container if necessary
        :param str parameter_server_registry_url: Registry server URL for accessing private docker registry container if necessary
        :param list[str] tags: List of tags

        :returns: experiment handle
        :rtype: str
        """
        self._validate_arguments(model_type=model_type, model_path=model_path)

        experiment_type_id = self._get_experiment_type_id(experiment_type_id)

        if not is_preemptible:
            is_preemptible = None

        datasets = self._dataset_dicts_to_instances(datasets)

        experiment = models.MultiNodeExperiment(
            name=name,
            project_id=project_id,
            experiment_type_id=experiment_type_id,
            worker_container=worker_container,
            worker_machine_type=worker_machine_type,
            worker_command=worker_command,
            worker_count=worker_count,
            parameter_server_container=parameter_server_container,
            parameter_server_machine_type=parameter_server_machine_type,
            parameter_server_command=parameter_server_command,
            parameter_server_count=parameter_server_count,
            ports=ports,
            workspace_url=workspace_url,
            workspace_ref=workspace_ref,
            workspace_username=workspace_username,
            workspace_password=workspace_password,
            datasets=datasets,
            working_directory=working_directory,
            artifact_directory=artifact_directory,
            cluster_id=cluster_id,
            experiment_env=experiment_env,
            model_type=model_type,
            model_path=model_path,
            is_preemptible=is_preemptible,
            worker_container_user=worker_container_user,
            worker_registry_username=worker_registry_username,
            worker_registry_password=worker_registry_password,
            worker_registry_url=worker_registry_url,
            parameter_server_container_user=parameter_server_container_user,
            parameter_server_registry_username=parameter_server_registry_username,
            parameter_server_registry_password=parameter_server_registry_password,
            parameter_server_registry_url=parameter_server_registry_url,
        )

        repository = self.build_repository(repositories.CreateMultiNodeExperiment)
        handle = repository.create(experiment)

        if tags:
            self.add_tags(entity_id=handle, tags=tags)
        return handle

    def create_mpi_multi_node(
            self,
            project_id,
            worker_container,
            worker_machine_type,
            worker_command,
            worker_count,
            name=None,
            master_container=None,
            master_machine_type=None,
            master_command=None,
            master_count=None,
            ports=None,
            workspace_url=None,
            workspace_ref=None,
            workspace_username=None,
            workspace_password=None,
            datasets=None,
            working_directory=None,
            artifact_directory=None,
            cluster_id=None,
            experiment_env=None,
            model_type=None,
            model_path=None,
            is_preemptible=False,
            worker_container_user=None,
            worker_registry_username=None,
            worker_registry_password=None,
            worker_registry_url=None,
            master_container_user=None,
            master_registry_username=None,
            master_registry_password=None,
            master_registry_url=None,
            tags=None,
    ):
        """
        Create multinode experiment using MPI

        :param str name: Name of new experiment  [required]
        :param str project_id: Project ID  [required]
        :param str worker_container: Worker container (dockerfile) [required]
        :param str worker_machine_type: Worker machine type  [required]
        :param str worker_command: Worker command  [required]
        :param int worker_count: Worker count  [required]
        :param str master_container: Master container  [required]
        :param str master_machine_type: Master machine type  [required]
        :param str master_command: Master command  [required]
        :param int master_count: Master count  [required]
        :param str ports: Port to use in new experiment
        :param str workspace_url: Project git repository url
        :param str workspace_ref: Git commit hash, branch name or tag
        :param str workspace_username: Project git repository username
        :param str workspace_password: Project git repository password
        :param dict|list[dict]|tuple[dict] datasets: Dict or list of dicts describing dataset(s) used in experiment.
                                                     Required keys: "url"
                                                     Optional keys: "tag" for S3 tag and "auth" for S3 token
        :param str working_directory: Working directory for the experiment
        :param str artifact_directory: Artifacts directory
        :param str cluster_id: Cluster ID
        :param dict experiment_env: Environment variables in a JSON
        :param str model_type: defines the type of model that is being generated by the experiment. Model type must be one of Tensorflow, ONNX, or Custom
        :param str model_path: Model path
        :param bool is_preemptible: Is preemptible
        :param str worker_container_user: Worker container user
        :param str worker_registry_username: Registry username for accessing private docker registry container if necessary
        :param str worker_registry_password: Registry password for accessing private docker registry container if necessary
        :param str worker_registry_url: Registry server URL for accessing private docker registry container if necessary
        :param str master_container_user: Master container user
        :param str master_registry_username: Registry username for accessing private docker registry container if necessary
        :param str master_registry_password: Registry password for accessing private docker registry container if necessary
        :param str master_registry_url: Registry server URL for accessing private docker registry container if necessary
        :param list[str] tags: List of tags

        :returns: experiment handle
        :rtype: str
        """
        self._validate_arguments(model_type=model_type, model_path=model_path)

        if not is_preemptible:
            is_preemptible = None

        datasets = self._dataset_dicts_to_instances(datasets)
        experiment_type_id = constants.ExperimentType.MPI_MULTI_NODE

        experiment = models.MpiMultiNodeExperiment(
            name=name,
            project_id=project_id,
            experiment_type_id=experiment_type_id,
            worker_container=worker_container,
            worker_machine_type=worker_machine_type,
            worker_command=worker_command,
            worker_count=worker_count,
            master_container=master_container,
            master_machine_type=master_machine_type,
            master_command=master_command,
            master_count=master_count,
            ports=ports,
            workspace_url=workspace_url,
            workspace_ref=workspace_ref,
            workspace_username=workspace_username,
            workspace_password=workspace_password,
            datasets=datasets,
            working_directory=working_directory,
            artifact_directory=artifact_directory,
            cluster_id=cluster_id,
            experiment_env=experiment_env,
            model_type=model_type,
            model_path=model_path,
            is_preemptible=is_preemptible,
            worker_container_user=worker_container_user,
            worker_registry_username=worker_registry_username,
            worker_registry_password=worker_registry_password,
            worker_registry_url=worker_registry_url,
            master_container_user=master_container_user,
            master_registry_username=master_registry_username,
            master_registry_password=master_registry_password,
            master_registry_url=master_registry_url,
        )

        repository = self.build_repository(repositories.CreateMpiMultiNodeExperiment)
        handle = repository.create(experiment)

        if tags:
            self.add_tags(entity_id=handle, tags=tags)
        return handle

    def run_single_node(
            self,
            project_id,
            machine_type,
            command,
            name=None,
            ports=None,
            workspace_url=None,
            workspace_ref=None,
            workspace_username=None,
            workspace_password=None,
            datasets=None,
            working_directory=None,
            artifact_directory=None,
            cluster_id=None,
            experiment_env=None,
            model_type=None,
            model_path=None,
            is_preemptible=False,
            container=None,
            container_user=None,
            registry_username=None,
            registry_password=None,
            registry_url=None,
            tags=None,
    ):
        """Create and start single node experiment

        :param str name: Name of new experiment  [required]
        :param str project_id: Project ID  [required]
        :param str machine_type: Machine type [required]
        :param str command: Container entrypoint command  [required]
        :param str ports: Port to use in new experiment
        :param str workspace_url: Project git repository url
        :param str workspace_ref: Git commit hash, branch name or tag
        :param str workspace_username: Project git repository username
        :param str workspace_password: Project git repository password
        :param dict|list[dict]|tuple[dict] datasets: Dict or list of dicts describing dataset(s) used in experiment.
                                                     Required keys: "url"
                                                     Optional keys: "tag" for S3 tag and "auth" for S3 token
        :param str working_directory: Working directory for the experiment
        :param str artifact_directory: Artifacts directory
        :param str cluster_id: Cluster ID
        :param dict experiment_env: Environment variables in a JSON
        :param str model_type: defines the type of model that is being generated by the experiment. Model type must be one of Tensorflow, ONNX, or Custom
        :param str model_path: Model path
        :param bool is_preemptible: Is preemptible
        :param str container: Container (dockerfile) [required]
        :param str container_user: Container user for running the specified command in the container. If no containerUser is specified, the user will default to 'root' in the container.
        :param str registry_username: Registry username for accessing private docker registry container if necessary
        :param str registry_password: Registry password for accessing private docker registry container if necessary
        :param str registry_url: Registry server URL for accessing private docker registry container if necessary
        :param list[str] tags: List of tags

        :returns: experiment handle
        :rtype: str
        """

        self._validate_arguments(model_type=model_type, model_path=model_path)

        if not is_preemptible:
            is_preemptible = None

        datasets = self._dataset_dicts_to_instances(datasets)

        experiment = models.SingleNodeExperiment(
            experiment_type_id=constants.ExperimentType.SINGLE_NODE,
            name=name,
            project_id=project_id,
            machine_type=machine_type,
            ports=ports,
            workspace_url=workspace_url,
            workspace_ref=workspace_ref,
            workspace_username=workspace_username,
            workspace_password=workspace_password,
            datasets=datasets,
            working_directory=working_directory,
            artifact_directory=artifact_directory,
            cluster_id=cluster_id,
            experiment_env=experiment_env,
            model_type=model_type,
            model_path=model_path,
            is_preemptible=is_preemptible,
            container=container,
            command=command,
            container_user=container_user,
            registry_username=registry_username,
            registry_password=registry_password,
            registry_url=registry_url,
        )

        repository = self.build_repository(repositories.RunSingleNodeExperiment)
        handle = repository.create(experiment)

        if tags:
            self.add_tags(entity_id=handle, tags=tags)
        return handle

    def run_multi_node(
            self,
            project_id,
            worker_container,
            worker_machine_type,
            worker_command,
            worker_count,
            parameter_server_container,
            parameter_server_machine_type,
            parameter_server_command,
            parameter_server_count,
            name=None,
            experiment_type_id=constants.ExperimentType.GRPC_MULTI_NODE,
            ports=None,
            workspace_url=None,
            workspace_ref=None,
            workspace_username=None,
            workspace_password=None,
            datasets=None,
            working_directory=None,
            artifact_directory=None,
            cluster_id=None,
            experiment_env=None,
            model_type=None,
            model_path=None,
            is_preemptible=False,
            worker_container_user=None,
            worker_registry_username=None,
            worker_registry_password=None,
            worker_registry_url=None,
            parameter_server_container_user=None,
            parameter_server_registry_username=None,
            parameter_server_registry_password=None,
            parameter_server_registry_url=None,
            tags=None,
    ):
        """Create and start multinode experiment

        :param str name: Name of new experiment  [required]
        :param str project_id: Project ID  [required]
        :param str worker_container: Worker container (dockerfile) [required]
        :param str worker_machine_type: Worker machine type  [required]
        :param str worker_command: Worker command  [required]
        :param int worker_count: Worker count  [required]
        :param str parameter_server_container: Parameter server container  [required]
        :param str parameter_server_machine_type: Parameter server machine type  [required]
        :param str parameter_server_command: Parameter server command  [required]
        :param int parameter_server_count: Parameter server count  [required]
        :param int|str experiment_type_id: Experiment Type ID [required]
        :param str ports: Port to use in new experiment
        :param str workspace_url: Project git repository url
        :param str workspace_ref: Git commit hash, branch name or tag
        :param str workspace_username: Project git repository username
        :param str workspace_password: Project git repository password
        :param dict|list[dict]|tuple[dict] datasets: Dict or list of dicts describing dataset(s) used in experiment.
                                                     Required keys: "url"
                                                     Optional keys: "tag" for S3 tag and "auth" for S3 token
        :param str working_directory: Working directory for the experiment
        :param str artifact_directory: Artifacts directory
        :param str cluster_id: Cluster ID
        :param dict experiment_env: Environment variables in a JSON
        :param str model_type: defines the type of model that is being generated by the experiment. Model type must be one of Tensorflow, ONNX, or Custom
        :param str model_path: Model path
        :param bool is_preemptible: Is preemptible
        :param str worker_container_user: Worker container user
        :param str worker_registry_username: Registry username for accessing private docker registry container if necessary
        :param str worker_registry_password: Registry password for accessing private docker registry container if necessary
        :param str worker_registry_url: Registry server URL for accessing private docker registry container if necessary
        :param str parameter_server_container_user: Parameter server container user
        :param str parameter_server_registry_username: Registry username for accessing private docker registry container if necessary
        :param str parameter_server_registry_password: Registry password for accessing private docker registry container if necessary
        :param str parameter_server_registry_url: Registry server URL for accessing private docker registry container if necessary
        :param list[str] tags: List of tags

        :returns: experiment handle
        :rtype: str
        """
        self._validate_arguments(model_type=model_type, model_path=model_path)

        experiment_type_id = self._get_experiment_type_id(experiment_type_id)

        if not is_preemptible:
            is_preemptible = None

        datasets = self._dataset_dicts_to_instances(datasets)

        experiment = models.MultiNodeExperiment(
            name=name,
            project_id=project_id,
            experiment_type_id=experiment_type_id,
            worker_container=worker_container,
            worker_machine_type=worker_machine_type,
            worker_command=worker_command,
            worker_count=worker_count,
            parameter_server_container=parameter_server_container,
            parameter_server_machine_type=parameter_server_machine_type,
            parameter_server_command=parameter_server_command,
            parameter_server_count=parameter_server_count,
            ports=ports,
            workspace_url=workspace_url,
            workspace_ref=workspace_ref,
            workspace_username=workspace_username,
            workspace_password=workspace_password,
            datasets=datasets,
            working_directory=working_directory,
            artifact_directory=artifact_directory,
            cluster_id=cluster_id,
            experiment_env=experiment_env,
            model_type=model_type,
            model_path=model_path,
            is_preemptible=is_preemptible,
            worker_container_user=worker_container_user,
            worker_registry_username=worker_registry_username,
            worker_registry_password=worker_registry_password,
            worker_registry_url=worker_registry_url,
            parameter_server_container_user=parameter_server_container_user,
            parameter_server_registry_username=parameter_server_registry_username,
            parameter_server_registry_password=parameter_server_registry_password,
            parameter_server_registry_url=parameter_server_registry_url,
        )

        repository = self.build_repository(repositories.RunMultiNodeExperiment)
        handle = repository.create(experiment)

        if tags:
            self.add_tags(entity_id=handle, tags=tags)

        return handle

    def run_mpi_multi_node(
            self,
            project_id,
            worker_container,
            worker_machine_type,
            worker_command,
            worker_count,
            master_container,
            master_machine_type,
            master_command,
            master_count,
            name=None,
            ports=None,
            workspace_url=None,
            workspace_ref=None,
            workspace_username=None,
            workspace_password=None,
            datasets=None,
            working_directory=None,
            artifact_directory=None,
            cluster_id=None,
            experiment_env=None,
            model_type=None,
            model_path=None,
            is_preemptible=False,
            worker_container_user=None,
            worker_registry_username=None,
            worker_registry_password=None,
            worker_registry_url=None,
            master_container_user=None,
            master_registry_username=None,
            master_registry_password=None,
            master_registry_url=None,
            tags=None,
    ):
        """Create and start multinode experiment using MPI

        :param str name: Name of new experiment  [required]
        :param str project_id: Project ID  [required]
        :param str worker_container: Worker container (dockerfile) [required]
        :param str worker_machine_type: Worker machine type  [required]
        :param str worker_command: Worker command  [required]
        :param int worker_count: Worker count  [required]
        :param str master_container: Master container  [required]
        :param str master_machine_type: Master machine type  [required]
        :param str master_command: Master command  [required]
        :param int master_count: Master count  [required]
        :param str ports: Port to use in new experiment
        :param str workspace_url: Project git repository url
        :param str workspace_ref: Git commit hash, branch name or tag
        :param str workspace_username: Project git repository username
        :param str workspace_password: Project git repository password
        :param dict|list[dict]|tuple[dict] datasets: Dict or list of dicts describing dataset(s) used in experiment.
                                                     Required keys: "url"
                                                     Optional keys: "tag" for S3 tag and "auth" for S3 token
        :param str working_directory: Working directory for the experiment
        :param str artifact_directory: Artifacts directory
        :param str cluster_id: Cluster ID
        :param dict experiment_env: Environment variables in a JSON
        :param str model_type: defines the type of model that is being generated by the experiment. Model type must be one of Tensorflow, ONNX, or Custom
        :param str model_path: Model path
        :param bool is_preemptible: Is preemptible
        :param str worker_container_user: Worker container user
        :param str worker_registry_username: Registry username for accessing private docker registry container if necessary
        :param str worker_registry_password: Registry password for accessing private docker registry container if necessary
        :param str worker_registry_url: Registry server URL for accessing private docker registry container if necessary
        :param str master_container_user: Master container user
        :param str master_registry_username: Registry username for accessing private docker registry container if necessary
        :param str master_registry_password: Registry password for accessing private docker registry container if necessary
        :param str master_registry_url: Registry server URL for accessing private docker registry container if necessary
        :param list[str] tags: List of tags

        :returns: experiment handle
        :rtype: str
        """
        self._validate_arguments(model_type=model_type, model_path=model_path)

        if not is_preemptible:
            is_preemptible = None

        datasets = self._dataset_dicts_to_instances(datasets)

        experiment_type_id = constants.ExperimentType.MPI_MULTI_NODE

        experiment = models.MpiMultiNodeExperiment(
            name=name,
            project_id=project_id,
            experiment_type_id=experiment_type_id,
            worker_container=worker_container,
            worker_machine_type=worker_machine_type,
            worker_command=worker_command,
            worker_count=worker_count,
            master_container=master_container,
            master_machine_type=master_machine_type,
            master_command=master_command,
            master_count=master_count,
            ports=ports,
            workspace_url=workspace_url,
            workspace_ref=workspace_ref,
            workspace_username=workspace_username,
            workspace_password=workspace_password,
            datasets=datasets,
            working_directory=working_directory,
            artifact_directory=artifact_directory,
            cluster_id=cluster_id,
            experiment_env=experiment_env,
            model_type=model_type,
            model_path=model_path,
            is_preemptible=is_preemptible,
            worker_container_user=worker_container_user,
            worker_registry_username=worker_registry_username,
            worker_registry_password=worker_registry_password,
            worker_registry_url=worker_registry_url,
            master_container_user=master_container_user,
            master_registry_username=master_registry_username,
            master_registry_password=master_registry_password,
            master_registry_url=master_registry_url,
        )

        repository = self.build_repository(repositories.RunMpiMultiNodeExperiment)
        handle = repository.create(experiment)

        if tags:
            self.add_tags(entity_id=handle, tags=tags)

        return handle

    def start(self, experiment_id):
        """Start existing experiment that has not run

        :param str experiment_id: Experiment ID

        :raises: exceptions.GradientSdkError
        """

        repository = self.build_repository(repositories.StartExperiment)
        repository.start(experiment_id)

    def stop(self, experiment_id):
        """Stop running experiment

        :param str experiment_id: Experiment ID

        :raises: exceptions.GradientSdkError
        """

        repository = self.build_repository(repositories.StopExperiment)
        repository.stop(experiment_id)

    def list(self, project_id=None, offset=None, limit=None, get_meta=False, tags=None):
        """Get a list of experiments. Optionally filter by project ID

        :param str|list|None project_id:
        :param int offset:
        :param int limit:
        :param bool get_meta: get dict of metadata like number of total items, etc. Setting to True changes rtype
        :param list[str]|tuple[str] tags: tags to filter with OR

        :return: experiments
        :rtype: list[models.SingleNodeExperiment|models.MultiNodeExperiment]|tuple[list[models.SingleNodeExperiment|models.MultiNodeExperiment],dict]
        """

        repository = self.build_repository(repositories.ListExperiments)
        experiments = repository.list(project_id=project_id, limit=limit, offset=offset, get_meta=get_meta, tags=tags)
        return experiments

    def get(self, experiment_id):
        """Get experiment instance

        :param str experiment_id: Experiment ID
        :rtype: models.SingleNodeExperiment|models.MultiNodeExperiment|MpiMultiNodeExperiment
        """
        repository = self.build_repository(repositories.GetExperiment)
        experiment = repository.get(experiment_id=experiment_id)
        return experiment

    def logs(self, experiment_id, line=1, limit=10000):
        """Show list of latest logs from the specified experiment.

        :param str experiment_id: Experiment ID
        :param int line: line number at which logs starts to display on screen
        :param int limit: maximum lines displayed on screen, default set to 10 000

        :returns: list of LogRows
        :rtype: list[models.LogRow]
        """

        repository = self.build_repository(repositories.ListExperimentLogs)
        logs = repository.list(id=experiment_id, line=line, limit=limit)
        return logs

    def yield_logs(self, experiment_id, line=1, limit=10000):
        """Get log generator. Polls the API for new logs

        :param str experiment_id:
        :param int line: line number at which logs starts to display on screen
        :param int limit: maximum lines displayed on screen, default set to 10 000

        :returns: generator yielding LogRow instances
        :rtype: Iterator[models.LogRow]
        """

        repository = self.build_repository(repositories.ListExperimentLogs)
        logs_generator = repository.yield_logs(id=experiment_id, line=line, limit=limit)
        return logs_generator

    def delete(self, experiment_id):
        repository = self.build_repository(repositories.DeleteExperiment)
        repository.delete(experiment_id)

    def _validate_arguments(self, **kwargs):
        if kwargs.get("model_path") and not kwargs.get("model_type"):
            raise InvalidParametersError(
                EXPERIMENT_MODEL_PATH_VALIDATION_ERROR
            )

    def get_metrics(self, experiment_id, start=None, end=None, interval="30s", built_in_metrics=None):
        """Get experiment metrics

        :param str experiment_id: ID of experiment
        :param datetime.datetime|str start:
        :param datetime.datetime|str end:
        :param str interval:
        :param list[str] built_in_metrics: List of metrics to get if different than default
                    Available builtin metrics: cpuPercentage, memoryUsage, gpuMemoryFree, gpuMemoryUsed, gpuPowerDraw,
                                            gpuTemp, gpuUtilization, gpuMemoryUtilization

        :returns: Metrics of and experiment
        :rtype: dict[str,dict[str,list[dict]]]
        """

        repository = self.build_repository(repositories.GetExperimentMetrics)
        metrics = repository.get(
            id=experiment_id,
            start=start,
            end=end,
            interval=interval,
            built_in_metrics=built_in_metrics,
        )
        return metrics

    def list_metrics(self, experiment_id, start=None, end=None, interval="30s"):
        """List experiment metrics

        :param str experiment_id: ID of experiment
        :param datetime.datetime|str start:
        :param datetime.datetime|str end:
        :param str interval:
        :returns: Metrics of and experiment
        :rtype: dict[str,dict[str,list[dict]]]
        """

        repository = self.build_repository(repositories.ListExperimentMetrics)
        metrics = repository.get(
            id=experiment_id,
            start=start,
            end=end,
            interval=interval,
        )
        return metrics


    def stream_metrics(self, experiment_id, interval="30s", built_in_metrics=None):
        """Stream live experiment metrics

        :param str experiment_id: ID of experiment
        :param str interval:
        :param list[str] built_in_metrics: List of metrics to get if different than default
                    Available builtin metrics: cpuPercentage, memoryUsage, gpuMemoryFree, gpuMemoryUsed, gpuPowerDraw,
                                            gpuTemp, gpuUtilization, gpuMemoryUtilization

        :returns: Generator object yielding live experiment metrics
        :rtype: Iterable[dict]
        """

        repository = self.build_repository(repositories.StreamExperimentMetrics)
        metrics = repository.stream(
            id=experiment_id,
            interval=interval,
            built_in_metrics=built_in_metrics,
        )
        return metrics

    def _dataset_dicts_to_instances(self, datasets):
        if not datasets:
            return None

        if isinstance(datasets, dict):
            datasets = [datasets]

        for ds in datasets:
            if not (ds.get("id") or ds.get("uri")):
                raise ResourceCreatingDataError("Error while creating experiment with dataset: "
                                                'id or uri key is required')

        for ds in datasets:
            volume_options = ds.setdefault("volume_options", {})
            volume_options.setdefault("kind", ds.pop("volume_kind", None))
            volume_options.setdefault("size", ds.pop("volume_size", None))

        datasets = [models.ExperimentDataset(**ds) for ds in datasets]
        return datasets



import marshmallow as ma

from .. import models
from ..serializers.base import BaseSchema
from .dataset_version import DatasetVersionSchema
from .storage_provider import StorageProviderSchema


class DatasetSchema(BaseSchema):
    MODEL = models.Dataset

    id = ma.fields.Str()
    name = ma.fields.Str()
    description = ma.fields.Str()
    storage_provider_id = ma.fields.Str(dump_to="storageProviderId", dump_only=True)
    storage_provider = ma.fields.Nested(StorageProviderSchema, load_from="storageProvider", load_only=True)


class DatasetRefSchema(DatasetSchema):
    MODEL = models.DatasetRef

    version = ma.fields.Nested(DatasetVersionSchema, load_only=True)



import copy
import marshmallow

from .base import BaseSchema
from .. import models, utils


class NotebookSchema(BaseSchema):
    MODEL = models.Notebook

    id = marshmallow.fields.Str()
    machine_type = marshmallow.fields.Str(load_from="machineType", dump_to="machineType")
    vm_type_id = marshmallow.fields.Int(load_from="vmTypeId", dump_to="vmTypeId")
    vm_type_label = marshmallow.fields.Str(load_from="vmTypeLabel", dump_to="vmTypeLabel")
    container_id = marshmallow.fields.Int(load_from="containerId", dump_to="containerId", allow_none=True)
    container = marshmallow.fields.Str(load_from="container", dump_to="container", allow_none=True)
    container_name = marshmallow.fields.Str(load_from="containerName", dump_to="containerName", allow_none=True)
    name = marshmallow.fields.Str()
    cluster_id = marshmallow.fields.Str(load_from="clusterId", dump_to="clusterId")
    registry_username = marshmallow.fields.Str(load_from="registryUsername", dump_to="registryUsername")
    registry_password = marshmallow.fields.Str(load_from="registryPassword", dump_to="registryPassword")
    command = marshmallow.fields.Str(load_from="command", dump_to="command")
    default_entrypoint = marshmallow.fields.Str(load_from="defaultEntrypoint", dump_to="defaultEntrypoint")
    container_user = marshmallow.fields.Str(load_from="containerUser", dump_to="containerUser")
    shutdown_timeout = marshmallow.fields.Int(load_from="shutdownTimeout", dump_to="shutdownTimeout")
    is_preemptible = marshmallow.fields.Bool(load_from="isPreemptible", dump_to="isPreemptible")
    is_public = marshmallow.fields.Bool(load_from="isPublic", dump_to="isPublic")
    project_id = marshmallow.fields.Str(load_from="projectId", dump_to="projectId")
    project_handle = marshmallow.fields.Str(load_from="projectHandle", dump_to="projectHandle")
    state = marshmallow.fields.Str()
    token = marshmallow.fields.Str()
    job_error = marshmallow.fields.Str(load_from="jobError", dump_to="jobError")
    job_handle = marshmallow.fields.Str(load_from="jobHandle", dump_to="jobHandle")
    container = marshmallow.fields.Str()
    container_url = marshmallow.fields.Str(load_from="containerUrl", dump_to="containerUrl")
    base_container = marshmallow.fields.Str(load_from="baseContainer", dump_to="baseContainerUrl")
    base_container_url = marshmallow.fields.Str(load_from="baseContainerUrl", dump_to="baseContainerUrl")
    workspace = marshmallow.fields.Str(load_from="workspace", dump_to="workspace")
    workspace_username = marshmallow.fields.Str(load_from="workspaceUsername", dump_to="workspaceUsername")
    workspace_password = marshmallow.fields.Str(load_from="workspacePassword", dump_to="workspacePassword")
    workspace_ref = marshmallow.fields.Str(load_from="workspaceRef", dump_to="workspaceRef")
    vm_type = marshmallow.fields.Str(load_from="vmType", dump_to="vmType")
    fqdn = marshmallow.fields.Str()
    namespace = marshmallow.fields.Str()
    tags = marshmallow.fields.List(marshmallow.fields.Str(), load_only=True)
    metrics_url = marshmallow.fields.Str(dump_to="metricsURL", load_from="metricsURL")
    dt_created = marshmallow.fields.DateTime(dump_to="dtCreated", load_from="dtCreated")
    dt_modified = marshmallow.fields.DateTime(dump_to="dtModified", load_from="dtModified")
    dt_started = marshmallow.fields.DateTime(dump_to="dtStarted", load_from="dtStarted")
    dt_finished = marshmallow.fields.DateTime(dump_to="dtFinished", load_from="dtFinished")
    dt_provisioning_started = marshmallow.fields.Str(dump_to="dtProvisioningStarted", load_from="dtProvisioningStarted")
    dt_provisioning_finished = marshmallow.fields.Str(dump_to="dtProvisioningFinished",
                                                      load_from="dtProvisioningFinished")
    dt_teardown_started = marshmallow.fields.DateTime(dump_to="dtTeardownStarted", load_from="dtTeardownStarted")
    dt_teardown_finished = marshmallow.fields.DateTime(dump_to="dtTeardownStarted", load_from="dtTeardownStarted")
    dt_workspace_upload_finished = marshmallow.fields.DateTime(dump_to="dtWorkspaceUploadFinished", load_from="dtWorkspaceUploadFinished")
    dt_deleted = marshmallow.fields.DateTime(dump_to="dtDeleted", load_from="dtDeleted")
    environment = marshmallow.fields.Dict(dump_to="environment", load_from="environment")

    @marshmallow.pre_dump
    def preprocess(self, data, **kwargs):
        data = copy.copy(data)

        utils.base64_encode_attribute(data, "command")
        return data


class NotebookStartSchema(BaseSchema):
    MODEL = models.NotebookStart

    notebook_id = marshmallow.fields.Str(load_from="notebookId", dump_to="notebookId")
    machine_type = marshmallow.fields.Str(load_from="machineType", dump_to="machineType")
    vm_type_id = marshmallow.fields.Int(load_from="vmTypeId", dump_to="vmTypeId")
    vm_type_label = marshmallow.fields.Str(load_from="vmTypeLabel", dump_to="vmTypeLabel")
    name = marshmallow.fields.Str(load_from="notebookName", dump_to="notebookName")
    cluster_id = marshmallow.fields.Str(load_from="clusterId", dump_to="clusterId")
    shutdown_timeout = marshmallow.fields.Int(load_from="shutdownTimeout", dump_to="shutdownTimeout")
    is_preemptible = marshmallow.fields.Bool(load_from="isPreemptible", dump_to="isPreemptible")
